@@ -1,7 +1,11 @@
 from __future__ import annotations
 
+import datetime
+import hashlib
+import json
 import os
 import shutil
+import sys
 import textwrap
 from argparse import Namespace
 from collections import defaultdict
@@ -9,18 +13,17 @@ from itertools import chain
 from pathlib import Path
 from typing import Iterable, Mapping, Sequence, cast
 
-import atoml
 import click
+import tomlkit
 from resolvelib.reporters import BaseReporter
 from resolvelib.resolvers import ResolutionImpossible, ResolutionTooDeep, Resolver
 
-from pdm import termui
+from pdm import signals, termui
 from pdm.cli.utils import (
     check_project_file,
     find_importable_files,
     format_lockfile,
     format_resolution_impossible,
-    frozen_requirement_from_dist,
     get_dist_location,
     merge_dictionary,
     save_version_specifiers,
@@ -30,6 +33,7 @@ from pdm.cli.utils import (
 from pdm.exceptions import NoPythonVersion, PdmUsageError, ProjectError
 from pdm.formats import FORMATS
 from pdm.formats.base import array_of_inline_tables, make_array, make_inline_table
+from pdm.models.caches import JSONFileCache
 from pdm.models.candidates import Candidate
 from pdm.models.python import PythonInfo
 from pdm.models.requirements import Requirement, parse_requirement, strip_extras
@@ -47,9 +51,13 @@ def do_lock(
     tracked_names: Iterable[str] | None = None,
     requirements: list[Requirement] | None = None,
     dry_run: bool = False,
+    refresh: bool = False,
 ) -> dict[str, Candidate]:
     """Performs the locking process and update lockfile."""
     check_project_file(project)
+    if refresh:
+        project.write_lockfile(project.lockfile)
+        return project.lockfile
     # TODO: multiple dependency definitions for the same package.
     provider = project.get_provider(strategy, tracked_names)
     if not requirements:
@@ -64,6 +72,7 @@ def do_lock(
         with ui.open_spinner(title="Resolving dependencies", spinner="dots") as spin:
             reporter = project.get_reporter(requirements, tracked_names, spin)
             resolver: Resolver = project.core.resolver_class(provider, reporter)
+            signals.pre_lock.send(project, requirements=requirements, dry_run=dry_run)
             try:
                 mapping, dependencies = resolve(
                     resolver,
@@ -84,10 +93,11 @@ def do_lock(
             except ResolutionImpossible as err:
                 spin.fail(f"{termui.Emoji.LOCK} Lock failed")
                 ui.echo(format_resolution_impossible(err), err=True)
-                raise
+                raise ResolutionImpossible("Unable to find a resolution") from None
             else:
-                data = format_lockfile(mapping, dependencies)
+                data = format_lockfile(project, mapping, dependencies)
                 spin.succeed(f"{termui.Emoji.LOCK} Lock successful")
+            signals.post_lock.send(project, resolution=mapping, dry_run=dry_run)
 
     project.write_lockfile(data, write=not dry_run)
 
@@ -118,6 +128,30 @@ def resolve_candidates_from_lockfile(
     return mapping
 
 
+def check_lockfile(project: Project, raise_not_exist: bool = True) -> str | None:
+    """Check if the lock file exists and is up to date. Return the update strategy."""
+    if not project.lockfile_file.exists():
+        if raise_not_exist:
+            raise ProjectError("Lock file does not exist, nothing to install")
+        project.core.ui.echo("Lock file does not exist", fg="yellow", err=True)
+        return "all"
+    elif not project.is_lockfile_compatible():
+        project.core.ui.echo(
+            "Lock file version is not compatible with PDM, installation may fail",
+            fg="yellow",
+            err=True,
+        )
+        return "all"
+    elif not project.is_lockfile_hash_match():
+        project.core.ui.echo(
+            "Lock file hash doesn't match pyproject.toml, packages may be outdated",
+            fg="yellow",
+            err=True,
+        )
+        return "reuse"
+    return None
+
+
 def do_sync(
     project: Project,
     *,
@@ -134,19 +168,6 @@ def do_sync(
 ) -> None:
     """Synchronize project"""
     if requirements is None:
-        if not project.lockfile_file.exists():
-            raise ProjectError("Lock file does not exist, nothing to sync")
-        elif not project.is_lockfile_compatible():
-            project.core.ui.echo(
-                "Lock file version is not compatible with PDM, "
-                "install may fail, please regenerate the pdm.lock",
-                err=True,
-            )
-        elif not project.is_lockfile_hash_match():
-            project.core.ui.echo(
-                "Lock file hash doesn't match pyproject.toml, packages may be outdated",
-                err=True,
-            )
         groups = translate_groups(project, default, dev, groups or ())
         requirements = []
         for group in groups:
@@ -163,10 +184,12 @@ def do_sync(
         dry_run,
         no_editable=no_editable,
         install_self=not no_self and "default" in groups and bool(project.meta.name),
-        use_install_cache=project.config["feature.install_cache"],
+        use_install_cache=project.config["install.cache"],
         reinstall=reinstall,
     )
+    signals.pre_install.send(project, candidates=candidates, dry_run=dry_run)
     handler.synchronize()
+    signals.post_install.send(project, candidates=candidates, dry_run=dry_run)
 
 
 def do_add(
@@ -182,6 +205,7 @@ def do_add(
     no_editable: bool = False,
     no_self: bool = False,
     dry_run: bool = False,
+    prerelease: bool = False,
 ) -> None:
     """Add packages and install"""
     check_project_file(project)
@@ -195,11 +219,12 @@ def do_add(
         parse_requirement(line) for line in packages
     ]:
         key = r.identify()
+        r.prerelease = prerelease
         tracked_names.add(key)
         requirements[key] = r
     project.core.ui.echo(
         f"Adding packages to {group} {'dev-' if dev else ''}dependencies: "
-        + ", ".join(termui.green(key or "", bold=True) for key in requirements)
+        + ", ".join(termui.green(r.as_line(), bold=True) for r in requirements.values())
     )
     all_dependencies = project.all_dependencies
     group_deps = all_dependencies.setdefault(group, {})
@@ -215,8 +240,7 @@ def do_add(
     save_version_specifiers({group: deps_to_update}, resolved, save)
     if not dry_run:
         project.add_dependencies(deps_to_update, group, dev)
-    lockfile = project.lockfile
-    project.write_lockfile(lockfile, False)
+        project.write_lockfile(project.lockfile, False)
 
     if sync:
         do_sync(
@@ -245,6 +269,7 @@ def do_update(
     sync: bool = True,
     no_editable: bool = False,
     no_self: bool = False,
+    prerelease: bool = False,
 ) -> None:
     """Update specified packages or all packages"""
     check_project_file(project)
@@ -257,6 +282,8 @@ def do_update(
     updated_deps: dict[str, dict[str, Requirement]] = defaultdict(dict)
     install_dev = True if dev is None else dev
     if not packages:
+        if prerelease:
+            raise PdmUsageError("--prerelease must be used with packages given")
         groups = translate_groups(project, default, install_dev, groups or ())
         for group in groups:
             updated_deps[group] = all_dependencies[group]
@@ -278,6 +305,7 @@ def do_update(
                         termui.green(name, bold=True), group, "dev-" if dev else ""
                     )
                 )
+            dependencies[matched_name].prerelease = prerelease
             updated_deps[group][matched_name] = dependencies[matched_name]
         project.core.ui.echo(
             "Updating packages: {}.".format(
@@ -299,7 +327,7 @@ def do_update(
         reqs,
         dry_run=dry_run,
     )
-    if sync:
+    if sync or dry_run:
         do_sync(
             project,
             groups=groups,
@@ -401,13 +429,16 @@ def do_list(
         if json:
             raise PdmUsageError("--json must be used with --graph")
         if freeze:
-            reqs = [
-                frozen_requirement_from_dist(dist)
-                for dist in sorted(
-                    working_set.values(), key=lambda d: d.metadata["Name"]
-                )
-            ]
-            project.core.ui.echo("".join(reqs))
+            reqs = sorted(
+                (
+                    Requirement.from_dist(dist).as_line()
+                    for dist in sorted(
+                        working_set.values(), key=lambda d: d.metadata["Name"]
+                    )
+                ),
+                key=lambda x: x.lower(),
+            )
+            project.core.ui.echo("\n".join(reqs))
             return
         rows = [
             (termui.green(k, bold=True), termui.yellow(v.version), get_dist_location(v))
@@ -436,6 +467,8 @@ def do_build(
         dest = project.root.joinpath(dest).as_posix()
     if clean:
         shutil.rmtree(dest, ignore_errors=True)
+    signals.pre_build.send(project, dest=dest, config_settings=config_settings)
+    artifacts: list[str] = []
     with project.core.ui.logging("build"):
         if sdist:
             project.core.ui.echo("Building sdist...")
@@ -443,12 +476,17 @@ def do_build(
                 dest, config_settings
             )
             project.core.ui.echo(f"Built sdist at {loc}")
+            artifacts.append(loc)
         if wheel:
             project.core.ui.echo("Building wheel...")
             loc = WheelBuilder(project.root, project.environment).build(
                 dest, config_settings
             )
             project.core.ui.echo(f"Built wheel at {loc}")
+            artifacts.append(loc)
+    signals.post_build.send(
+        project, artifacts=artifacts, config_settings=config_settings
+    )
 
 
 def do_init(
@@ -468,13 +506,13 @@ def do_init(
             "description": "",
             "authors": array_of_inline_tables([{"name": author, "email": email}]),
             "license": make_inline_table({"text": license}),
-            "urls": {"homepage": ""},
+            "urls": {"Homepage": ""},
             "dependencies": make_array([], True),
-            "requires-python": "" if python_requires == "*" else python_requires,
-            "dynamic": ["classifiers"],
         },
         "build-system": {"requires": ["pdm-pep517"], "build-backend": "pdm.pep517.api"},
     }
+    if python_requires and python_requires != "*":
+        data["project"]["requires-python"] = python_requires  # type: ignore
     get_specifier(python_requires)
     if not project.pyproject:
         project._pyproject = data
@@ -482,39 +520,71 @@ def do_init(
         project._pyproject["project"] = data["project"]  # type: ignore
         project._pyproject["build-system"] = data["build-system"]  # type: ignore
     project.write_pyproject()
+    signals.post_init.send(project)
 
 
-def do_use(project: Project, python: str = "", first: bool = False) -> None:
+def do_use(
+    project: Project,
+    python: str = "",
+    first: bool = False,
+    ignore_remembered: bool = False,
+) -> None:
     """Use the specified python version and save in project config.
     The python can be a version string or interpreter path.
     """
+    if python:
+        python = python.strip()
 
     def version_matcher(py_version: PythonInfo) -> bool:
         return project.python_requires.contains(str(py_version.version))
 
-    if python:
-        python = python.strip()
-
-    found_interpreters = list(
-        dict.fromkeys(filter(version_matcher, project.find_interpreters(python)))
+    if not project.cache_dir.exists():
+        project.cache_dir.mkdir(parents=True)
+    use_cache: JSONFileCache[str, str] = JSONFileCache(
+        project.cache_dir / "use_cache.json"
     )
-    if not found_interpreters:
-        raise NoPythonVersion("Python interpreter is not found on the system.")
-    if first or len(found_interpreters) == 1:
-        selected_python = found_interpreters[0]
-    else:
-        project.core.ui.echo("Please enter the Python interpreter to use")
-        for i, py_version in enumerate(found_interpreters):
-            project.core.ui.echo(
-                f"{i}. {termui.green(py_version.executable)} ({py_version.identifier})"
+    selected_python: PythonInfo | None = None
+    if python and not ignore_remembered:
+        if use_cache.has_key(python):
+            cached_python = PythonInfo.from_path(use_cache.get(python))
+            if version_matcher(cached_python):
+                project.core.ui.echo(
+                    "Using the last selection, add '-i' to ignore it.",
+                    fg="yellow",
+                    err=True,
+                )
+                selected_python = cached_python
+    if selected_python is None:
+        found_interpreters = list(dict.fromkeys(project.find_interpreters(python)))
+        matching_interperters = list(filter(version_matcher, found_interpreters))
+        if not found_interpreters:
+            raise NoPythonVersion("Python interpreter is not found on the system.")
+        if not matching_interperters:
+            project.core.ui.echo("Interpreters found but not matching:", err=True)
+            for py in found_interpreters:
+                project.core.ui.echo(f"  - {py.executable} ({py.identifier})", err=True)
+            raise NoPythonVersion(
+                "No python is found meeting the requirement "
+                f"{termui.green('python' + str(project.python_requires))}"
             )
-        selection = click.prompt(
-            "Please select:",
-            type=click.Choice([str(i) for i in range(len(found_interpreters))]),
-            default="0",
-            show_choices=False,
-        )
-        selected_python = found_interpreters[int(selection)]
+        if first or len(found_interpreters) == 1:
+            selected_python = found_interpreters[0]
+        else:
+            project.core.ui.echo("Please enter the Python interpreter to use")
+            for i, py_version in enumerate(found_interpreters):
+                project.core.ui.echo(
+                    f"{i}. {termui.green(py_version.executable)} "
+                    f"({py_version.identifier})"
+                )
+            selection = click.prompt(
+                "Please select:",
+                type=click.Choice([str(i) for i in range(len(found_interpreters))]),
+                default="0",
+                show_choices=False,
+            )
+            selected_python = found_interpreters[int(selection)]
+        if python:
+            use_cache.set(python, selected_python.path)
 
     old_python = project.python if "python.path" in project.config else None
     project.core.ui.echo(
@@ -560,18 +630,18 @@ def do_import(
     if options is None:
         options = Namespace(dev=False, group=None)
     project_data, settings = FORMATS[key].convert(project, filename, options)
-    pyproject = project.pyproject or atoml.document()
+    pyproject = project.pyproject or tomlkit.document()
 
     if "tool" not in pyproject or "pdm" not in pyproject["tool"]:  # type: ignore
-        pyproject.setdefault("tool", {})["pdm"] = atoml.table()
+        pyproject.setdefault("tool", {})["pdm"] = tomlkit.table()
 
     if "project" not in pyproject:
-        pyproject.add("project", atoml.table())  # type: ignore
+        pyproject.add("project", tomlkit.table())  # type: ignore
         pyproject["project"].add(  # type: ignore
-            atoml.comment("PEP 621 project metadata")
+            tomlkit.comment("PEP 621 project metadata")
         )
         pyproject["project"].add(  # type: ignore
-            atoml.comment("See https://www.python.org/dev/peps/pep-0621/")
+            tomlkit.comment("See https://www.python.org/dev/peps/pep-0621/")
         )
 
     merge_dictionary(pyproject["project"], project_data)  # type: ignore
@@ -678,56 +748,29 @@ def print_pep582_command(ui: termui.UI, shell: str = "AUTO") -> None:
     ui.echo(result)
 
 
-def migrate_pyproject(project: Project) -> None:
-    """Migrate the legacy pyproject format to PEP 621"""
+def get_latest_version(project: Project) -> str | None:
+    """Get the latest version of PDM from PyPI, cache for 7 days"""
+    from pdm.utils import get_finder
 
-    if project.pyproject and "project" in project.pyproject:
-        pyproject = project.pyproject
-        settings = {}
-        updated_fields = []
-        for field in ("includes", "excludes", "build", "package-dir"):
-            if field in pyproject["project"]:
-                updated_fields.append(field)
-                settings[field] = pyproject["project"].pop(field)
-        if "dev-dependencies" in pyproject["project"]:
-            if pyproject["project"]["dev-dependencies"]:
-                settings["dev-dependencies"] = {
-                    "dev": pyproject["project"]["dev-dependencies"]
-                }
-            del pyproject["project"]["dev-dependencies"]
-            updated_fields.append("dev-dependencies")
-        if updated_fields:
-            if "tool" not in pyproject or "pdm" not in pyproject["tool"]:
-                pyproject.setdefault("tool", {})["pdm"] = atoml.table()
-            pyproject["tool"]["pdm"].update(settings)
-            project.pyproject = pyproject
-            project.write_pyproject()
-            project.core.ui.echo(
-                f"{termui.yellow('[AUTO-MIGRATION]')} These fields are moved from "
-                f"[project] to [tool.pdm] table: {updated_fields}",
-                err=True,
-            )
-        return
-
-    if not project.pyproject_file.exists() or not FORMATS["legacy"].check_fingerprint(
-        project, project.pyproject_file
+    cache_key = hashlib.sha224(sys.executable.encode()).hexdigest()
+    cache_file = project.cache("self-check") / cache_key
+    if cache_file.exists():
+        state = json.loads(cache_file.read_text())
+    else:
+        state = {}
+    current_time = datetime.datetime.utcnow().timestamp()
+    if (
+        state.get("last-check")
+        and current_time - state["last-check"] < 60 * 60 * 24 * 7
     ):
-        return
-
-    project.core.ui.echo(
-        f"{termui.yellow('[AUTO-MIGRATION]')} Legacy pdm 0.x metadata detected, "
-        "migrating to PEP 621...",
-        err=True,
-    )
-    do_import(project, str(project.pyproject_file), "legacy")
-    project.core.ui.echo(
-        termui.green("pyproject.toml")
-        + termui.yellow(
-            " has been migrated to PEP 621 successfully. "
-            "Now you can safely delete the legacy metadata under [tool.pdm] table."
-        ),
-        err=True,
-    )
+        return cast(str, state["latest-version"])
+    candidate = get_finder([], project.cache_dir.as_posix()).find_best_candidate("pdm")
+    if not candidate.best_candidate:
+        return None
+    latest_version = str(candidate.best_candidate.version)
+    state.update({"latest-version": latest_version, "last-check": current_time})
+    cache_file.write_text(json.dumps(state))
+    return latest_version
 
 
 def check_update(project: Project) -> None:
@@ -742,14 +785,13 @@ def check_update(project: Project) -> None:
         is_pipx_installation,
         is_scoop_installation,
     )
-    from pdm.utils import get_finder
 
-    this_version = parse_version(project.core.version)
-    candidate = get_finder([]).find_best_candidate("pdm")
-    if not candidate.best_candidate or candidate.best_candidate.version <= this_version:
+    this_version = project.core.version
+    latest_version = get_latest_version(project)
+    if latest_version is None or parse_version(this_version) >= parse_version(
+        latest_version
+    ):
         return
-
-    latest_version = candidate.best_candidate.version
     if is_pipx_installation():  # pragma: no cover
         install_command = "$ pipx upgrade pdm"
     elif is_scoop_installation():  # pragma: no cover
@@ -762,8 +804,8 @@ def check_update(project: Project) -> None:
     disable_command = "$ pdm config check_update false"
 
     message = [
-        termui.blue(f"\nPDM {termui.cyan(str(this_version))}"),
-        termui.blue(f" is installed, while {termui.cyan(str(latest_version))}"),
+        termui.blue(f"\nPDM {termui.cyan(this_version)}"),
+        termui.blue(f" is installed, while {termui.cyan(latest_version)}"),
         termui.blue(" is available.\n"),
         termui.blue(f"Please run {termui.green(install_command, bold=True)}"),
         termui.blue(" to upgrade.\n"),
